import asyncio
import logging
import re
import traceback
from pathlib import Path
from typing import Any

import black
import isort
import tqdm
from pydantic import BaseModel, Field

from promptedgraphs.llms.coding import fix_code

logger = logging.getLogger(__name__)

# Restrictive environment
allowed_globals = {
    "__import__": __import__,
    "BaseModel": BaseModel,
    "Field": Field,
    "bytes": bytes,
}


def kindofsafe_exec(code, dependencies: dict = None):
    variables = {}
    dependencies = dependencies or allowed_globals

    for snippet in code.split(
        "\n\n"
    ):  # Loading in code by chunks helps with dependency resolution
        dependencies.update(variables)
        exec(snippet, dependencies, variables)
    return variables


def is_pydantic_base_model(cls: Any) -> bool:
    return isinstance(cls, type) and issubclass(cls, BaseModel)


def extract_fields_and_type(cls: type[BaseModel]) -> list[tuple[str, type, str]]:
    results = []
    if type(cls) == type(BaseModel):
        return []
    schema = cls.model_json_schema()
    properties = schema.get("properties", {})
    for field_name, field_info in properties.items():
        if field_info["type"] == "array":
            results.append((field_name, cls, "item"))
        else:
            results.append((field_name, cls, "object"))
    return results


def process_file(file_path):
    with open(file_path) as file:
        file_content = file.read()

    # Things can throw weird errors if multiple definitions of the same name are in the same file

    local_vars = kindofsafe_exec(file_content)
    for name, cls in local_vars.items():
        if issubclass(cls, BaseModel) and hasattr(cls, "model_json_schema"):
            cls.model_json_schema()

            # TODO  GO HAM ON THIS


def aggregate_python_files(fdir):
    fdir = Path(fdir)
    code = []

    for fname in fdir.glob("*.py"):
        with open(fname) as f:
            c = black.format_str(f.read(), mode=black.FileMode())
            code.append(c)
            try:
                black.format_str("\n\n".join(code), mode=black.FileMode())
            except Exception:
                logger.error(f"Static code error in: {fname}")
                return

    # move all imports to the top
    code = move_imports_to_top("\n\n".join(code))
    config = isort.Config(profile="black")
    code = isort.code(code, config=config)

    # reformat one more time
    code = black.format_str(code, mode=black.FileMode())

    with open(fdir / "_all.py", "w") as f:
        f.write(code)

    return fdir / "_all.py"


def move_imports_to_top(code):
    lines = code.split("\n")

    # Separating import lines and other lines
    import_lines = [
        line for line in lines if re.match(r"^(import |from .+ import)", line)
    ]
    other_lines = [
        line for line in lines if not re.match(r"^(import |from .+ import)", line)
    ]

    # Combining the lines and writing back to the file
    return "\n".join(import_lines + other_lines)


async def validate_python_files(fdir):
    fnames = sorted(fdir.glob("*.py"))
    for fname in tqdm.tqdm(fnames):
        if fname.name == "_all.py":
            continue
        code = Path(fname).read_text()

        history = []
        i = 0
        while i < 4:
            code = black.format_str(code, mode=black.FileMode())
            kindofsafe_exec(code)
            # break
            try:
                code = black.format_str(code, mode=black.FileMode())
                kindofsafe_exec(code)
                break
            except Exception as e:
                i += 1
                if i >= 4:
                    logger.error(f"Code error in: {fname}")
                    return

                logger.warning(f"fixing code error in: {fname} - take {i} - {e}")
                code, history = await fix_code(
                    code, error=e, tb=traceback.format_exc(), history=history
                )
        if i > 0:
            logger.warning(f"Fixed code error in: {fname}")
            with open(fname, "w") as f:
                f.write(code)


async def python_files_pipeline(fdir):
    await validate_python_files(fdir)
    file_path = aggregate_python_files(fdir)
    results = process_file(file_path)
    print(results)
    return results


if __name__ == "__main__":
    fdir = Path("data_models/googlemaps.client.Client")
    asyncio.run(python_files_pipeline(fdir))

    # for r in results:



import json
import re

from promptedgraphs.config import Config, load_config
from promptedgraphs.llms.openai_streaming import (
    GPT_MODEL,
    streaming_chat_completion_request,
)


async def _sync_wrapper(messages, config: Config | None = None, model=GPT_MODEL):
    payload = ""
    async for event in streaming_chat_completion_request(
        messages=messages,
        functions=None,
        config=config or load_config(),
        stream=False,
        model=model,
    ):
        if event.data:
            payload += event.data
        elif event.retry:
            print(f"Retry: {event.retry}")
    data = json.loads(payload)
    return data


def extract_code_blocks(text: str) -> list[dict[str, str]]:
    pattern = re.compile(r"```(.*?)\n(.*?)```", re.DOTALL)
    matches = pattern.findall(text)

    blocks = []
    for match in matches:
        block_type, content = match
        blocks.append({"block_type": block_type.strip(), "content": content.strip()})
    return blocks



import asyncio


async def async_generator_1():
    for i in range(5):
        await asyncio.sleep(1)
        yield f"Gen1: {i}"


async def async_generator_2():
    for i in range(5):
        await asyncio.sleep(2)
        yield f"Gen2: {i}"


async def merge_generators(*generators):
    tasks = {asyncio.create_task(gen()): gen for gen in generators}
    while tasks:
        done, _ = await asyncio.wait(tasks.keys(), return_when=asyncio.FIRST_COMPLETED)
        for task in done:
            if task.exception():
                continue
            yield task.result()
            # Restart the task
            gen = tasks.pop(task)
            tasks[asyncio.create_task(gen())] = gen


async def main():
    async for item in merge_generators(async_generator_1, async_generator_2):
        print(item)


asyncio.run(main())



# get traceback

from promptedgraphs.llms.helpers import _sync_wrapper, extract_code_blocks
from promptedgraphs.models import ChatMessage


async def fix_code(
    code: str, error: str, tb: str, history: list[tuple[str, str]]
) -> tuple[str, list[tuple[str, str]]]:
    """return new code plus history of code, tb pairs"""

    system_message = """You are an expert programmer working with a user to debug and fix code
    you will write the corrected code and the user provides you with traceback information
    after running your code on their system.

    A history of errors and code changes is shown in the conversation.  Please fix the code
    and respond with the corrected code in a code block"""

    history = history or []
    history.append((code, f"{error}\n\n{tb}" if error else tb or ""))

    messages = [
        ChatMessage(
            role="system",
            content=system_message,
        )
    ]
    for c, err in history:
        messages.append(
            ChatMessage(
                role="assistant",
                content=f"```python\n{c}\n```",
            )
        )
        messages.append(
            ChatMessage(
                role="user",
                content=f"# ERROR and STACKTRACE\n```bash\n{err}\n```",
            )
        )
    data = await _sync_wrapper(messages, model="gpt-4-0613")

    # pull out the json and pydantic code blocks
    content = data.get("choices")[0]["message"]["content"]
    results = extract_code_blocks(content)
    for r in results:
        if r["block_type"] == "python":
            return r["content"], history



"""Sometimes you just have to RTFM.
Occationally documentation contains links that are helpful to understand the APIs and usage.

This module contains functions to help you bootstrap a datagraph from APIs and functions.

In particular, given a url

 * renders the webpage (or portion of the webpage) with javascript enabled
 * formats the webpage as markdown
"""

import os
from urllib.parse import quote

import requests

from promptedgraphs.config import Config


class OGTags:
    def __init__(self, token: str = None, config: Config = None):
        if token:
            self.token = token
        elif config:
            self.token = config.ogtags_api_key
        else:
            self.token = os.environ.get("OGTAGS_API_KEY")
        assert (
            self.token
        ), "You must provide an API token to use OGTags, set OGTAGS_API_KEY in your environment"

    def get_page_source_via_chrome_render(self, crawl_url):
        encoded_url = quote(crawl_url, safe="")
        full_url = f"https://api.ogtags.dev/v1/page-source/chrome-render/{encoded_url}"

        headers = {
            "accept": "application/json",
            "Authorization": f"Bearer {self.token}",
        }
        params = {"width": "1200", "height": "630"}
        response = requests.get(full_url, headers=headers, params=params)

        if response.status_code != 200:
            raise RuntimeError(
                f"Request to OGTags failed with status code {response.status_code} and message {response.text}"
            )

        return response.text


def fetch_from_ogtags(crawl_url):
    ogtags = OGTags()
    return ogtags.get_page_source_via_chrome_render(crawl_url)


if __name__ == "__main__":
    # Example usage:
    crawl_url = (
        "https://developers.google.com/places/web-service/search#FindPlaceRequests"
    )
    result = fetch_from_ogtags(crawl_url)
    print(result)



import seaborn as sns
from pydantic import BaseModel
from spacy import displacy

from promptedgraphs.models import EntityReference


def rgb_to_hex(rgb):
    return "#{:02x}{:02x}{:02x}".format(
        int(rgb[0] * 256), int(rgb[1] * 256), int(rgb[2] * 256)
    )


def render_entities(
    text: str,
    ents: BaseModel | list[EntityReference] | dict = None,
    jupyter=True,
    color_dict: dict = None,
    color_palette: list[float] = None,
    **options
):
    """Renders entities using the displacy.render function"""

    if ents is None:
        return displacy.render(
            {"text": text}, style="ent", jupyter=jupyter, manual=True
        )
    elif hasattr(ents, "model_dump"):
        data = ents.model_dump()
        fields = sorted(data.keys())
        ents = [
            EntityReference(
                start=text.find(v), end=text.find(v) + len(v), label=k, text=v
            )
            for k, v in data.items()
            if v in text
        ]
    elif isinstance(ents, dict):
        data = ents
        fields = sorted(data.keys())
        ents = [
            EntityReference(
                start=text.find(v), end=text.find(v) + len(v), label=k, text=v
            )
            for k, v in data.items()
            if v in text
        ]
    elif isinstance(ents[0], EntityReference):
        fields = sorted({e.label for e in ents})
    else:
        raise ValueError("ents must be a list of EntityReference, BaseModel or dict")

    # Build colors
    if color_dict is None:
        palette = color_palette or sns.color_palette("Set2", 8)
        color_dict = {f: rgb_to_hex(color) for f, color in zip(list(fields), palette)}

    return displacy.render(
        {
            "text": text,
            "ents": [e.to_dict() for e in ents],
        },
        style="ent",
        jupyter=jupyter,
        manual=True,
        options={"colors": color_dict, **options},
    )



"""Bootstrap a datagraph from APIs and functions"""
import asyncio
import datetime
import inspect
import json
import logging
import os
import re
from pathlib import Path

import googlemaps
import tqdm
from bs4 import BeautifulSoup
from dotenv import load_dotenv

from promptedgraphs.config import load_config
from promptedgraphs.llms.helpers import _sync_wrapper, extract_code_blocks
from promptedgraphs.llms.openai_streaming import streaming_chat_completion_request
from promptedgraphs.models import ChatMessage
from promptedgraphs.sources.rtfm import fetch_from_ogtags

load_dotenv()
load_config()

logger = logging.getLogger(__name__)

SYSTEM_MESSAGE_FN_TO_SCHEMAS = """You are a python developer tasked with building pydantic data models to enforce types for various function calls.
Below is a function's call signature and doc string with optionally additional documentation

# Steps:
1.  Write the function as an openAPI spec in JSON assuming the endpoint is GET `/{fn_name}`.  Make sure to enclose the JSON data in a code block

2. Then convert the openAPI json spec above into pydantic data models.  Use pydantic.Field to keep types, title and description information.

Make sure to enclose the pydantic data models within a code block like
```python
from typing import Dict, List
from pydantic import BaseModel, Field


class ExampleRequest(BaseModel):
    pass


class ExampleResponse(BaseModel):
    pass
```

The response should only contain two code blocks (one with JSON data and one with python code)
"""


def get_functions_from_object(fn: classmethod):
    """Get functions from an object"""
    return [
        getattr(fn, a)
        for a in dir(fn)
        if callable(getattr(fn, a)) and not a.startswith("_")
    ]


def build_function_signature(fn: classmethod):
    """Build a function from a signature"""
    sig = inspect.signature(fn)
    doc = (fn.__doc__ or "").strip()
    return f'def {fn.__name__}{str(sig)}:\n    """{doc}' + '\n    """\n    ....'


def extract_urls_with_anchors(text):
    if not text:
        return []
    url_pattern = re.compile(
        r"http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+(?:#[a-zA-Z0-9-_]*)?"
    )
    return sorted(set(url_pattern.findall(text)))


def slim_down_html(html: str, anchor: str = None) -> tuple[str, bool]:
    html = BeautifulSoup(html, features="html.parser")
    # remove all script and style and visual elements
    for script in html(["script", "style", "svg", "img"]):
        script.extract()

    if anchor:
        if anchor := html.find(id=anchor):
            return str(anchor), True

    if article := html.find("article"):
        return str(article), False
    elif main := html.find("main"):
        return str(main), False
    elif body := html.find("body"):
        return str(body), False
    return str(html), False


async def html_to_markdown(html, url=None):
    """Convert html to markdown"""
    if url and "#" in url:
        anchor = url.split("#")[1] if url else None
    else:
        anchor = None
    html, used_anchor = slim_down_html(html, anchor=anchor)
    if anchor and not used_anchor:
        url = url.split("#")[0]

    SYSTEM_MESSAGE = """Format webpage as markdown"""
    payload = ""
    async for event in streaming_chat_completion_request(
        messages=[
            ChatMessage(
                role="system",
                content=SYSTEM_MESSAGE,
            ),
            ChatMessage(role="user", content=html),
        ],
        functions=None,
        config=load_config(),
        stream=False,
    ):
        if event.data:
            payload += event.data
        elif event.retry:
            print(f"Retry: {event.retry}")

    data = json.loads(payload)
    return data["choices"][0]["message"]["content"], url


async def build_function_schemas(
    fn_name: str, fn_signature: str, external_references: dict[str, str] = None
) -> tuple[dict]:
    """Convert function signatures into OpenAPI specs and pydantic datamodels"""

    external_references = external_references or {}

    messages = [
        ChatMessage(
            role="system",
            content=SYSTEM_MESSAGE_FN_TO_SCHEMAS.format(fn_name=fn_name),
        )
    ]
    messages.append(ChatMessage(role="user", content=fn_signature))
    messages.extend(
        ChatMessage(
            role="assistant",
            content=f"Additional documentation loaded from {url}\n\n{html}",
        )
        for url, html in external_references.items()
    )
    data = await _sync_wrapper(messages)

    # pull out the json and pydantic code blocks
    content = data.get("choices")[0]["message"]["content"]
    results = extract_code_blocks(content)
    if len(results) != 2:  # Try again
        missing_types = {"json", "python"} - {r["block_type"] for r in results}
        logger.warning(f"Missing code blocks of type: {missing_types}")
        messages.append(
            ChatMessage(
                role="assistant",
                content=content,
            )
        )
        messages.append(
            ChatMessage(
                role="user",
                content=f"Please provide the missing `{','.join(missing_types)}` code block",
            ),
        )
        data = await _sync_wrapper(messages)
        content = data.get("choices")[0]["message"]["content"]
        results.extend(extract_code_blocks(content))

    # Append the function signature documenation to the python code block
    docs = f"""# {fn_name}\n\n```python\n{fn_signature}\n\n```"""
    for url, html in external_references.items():
        docs += f"\n\n## Additional documentation loaded from {url}\n\n{html}"
    results.append({"block_type": "md", "content": docs})

    return results


async def register_function_as_datasource(
    obj: object,
    name: str = None,
    datasource_registry: str = None,
):
    name = (
        name or getattr(obj, "__name__", None) or str(obj).split()[0].replace("<", "")
    )
    assert name, "You must provide a name for the datasource"

    datasource_registry = Path(datasource_registry or "./data_models/")
    output_dir = datasource_registry / name
    output_dir.mkdir(exist_ok=True, parents=True)

    meta = {
        "name": name,
        "description": obj.__doc__ or "",
        "as_of": datetime.datetime.now(tz=datetime.timezone.utc).isoformat(),
    }

    fns = get_functions_from_object(obj)
    ittr = tqdm.tqdm(fns, desc=f"Building schemas for {name}")
    for fn in ittr:
        ittr.set_description(f"Building schema for {name}::{fn.__name__}")
        links = extract_urls_with_anchors(fn.__doc__)
        sig = build_function_signature(fn)

        external_references = {}
        for link in tqdm.tqdm(links, desc="Fetching external references"):
            if link in external_references:
                continue
            if html := fetch_from_ogtags(link):
                html, new_link = await html_to_markdown(html, url=link)
                external_references[new_link] = html

        fn_schemas = await build_function_schemas(
            fn.__name__, sig, external_references=external_references
        )

        for fn_schema in fn_schemas:
            suffix = (
                "py" if fn_schema["block_type"] == "python" else fn_schema["block_type"]
            )
            with open(output_dir / f"{fn.__name__}.{suffix}", "w") as f:
                f.write(fn_schema["content"])

    with open(datasource_registry / "meta.jsonl", "a") as f:
        f.write(json.dumps(meta) + "\n")

    # Aggregate all the schemas into one file


if __name__ == "__main__":
    load_dotenv()
    gmaps = googlemaps.Client(key=os.environ["GOOGLEMAPS_API_KEY"])
    asyncio.run(register_function_as_datasource(gmaps))



import asyncio
import contextlib
import json

from pydantic import BaseModel, Field

from promptedgraphs.config import Config
from promptedgraphs.data_extraction import (
    camelcase_to_words,
    create_functions,
    format_fieldinfo,
    remove_nas,
)
from promptedgraphs.llms.openai_streaming import (
    GPT_MODEL,
    streaming_chat_completion_request,
)
from promptedgraphs.models import ChatMessage
from promptedgraphs.parsers import extract_partial_list

BRAINSTORM_SYSTEM_MESSAGE = """
You are a Creative Director and Ideation Specialist. Your task is brainstorm examples of objects to be passed into python's `{name}(BaseModel)` pydantic class.
Generate {count} examples of objects that fit the following fields: {label_list}

{label_definitions}

{positive_examples}

{negative_examples}
"""


def create_messages(
    text,
    name,
    labels,
    n=10,
    positive_examples=None,
    negative_examples=None,
    custom_system_message=None,
    **kwargs,
):
    messages = []
    if custom_system_message:
        messages.append(
            ChatMessage(
                role="system",
                content=custom_system_message.strip(),
            )
        )
    else:
        label_list = list(labels.keys())

        label_definitions = """Below are definitions of each field to help aid you in what kinds of structured data to extract for each label.
    Assume these definitions are written by an expert and follow them closely.\n\n""" + "\n".join(
            [f" * {format_fieldinfo(k, dict(labels[k]))}" for k in label_list]
        )

        custom_system_message = BRAINSTORM_SYSTEM_MESSAGE.format(
            name=name,
            label_list=label_list,
            label_definitions=label_definitions,
            count=n,
            positive_examples="Good Examples:\n{positive_examples}\n"
            if positive_examples
            else "",
            negative_examples="Bad Examples:\n{negative_examples}\n"
            if negative_examples
            else "",
            **kwargs,
        )
        messages.append(
            ChatMessage(
                role="system",
                content=custom_system_message.strip(),
            )
        )
    if text:
        messages.append(ChatMessage(role="user", content=text))

    return messages


async def streaming_chat_wrapper(
    messages,
    functions,
    model,
    config,
    temperature,
    output_type,
    is_parent_list,
    name,
):
    count = 0
    payload = ""
    async for msg in streaming_chat_completion_request(
        messages=messages,
        functions=functions,
        model=model,
        config=config,
        temperature=temperature,
    ):
        if msg.data is None or msg.data == "":
            continue

        if msg.data == "[DONE]":
            with contextlib.suppress(json.decoder.JSONDecodeError):
                if is_parent_list:
                    s = json.loads(payload).get(name, [])
                    for data in s[count:]:
                        yield output_type(**remove_nas(data))
                else:
                    data = json.loads(payload)
                    yield output_type(**remove_nas(data))
            break

        if msg.event == "error":
            print("ERROR", msg.data)
            break

        data = json.loads(msg.data)

        choices = data.get("choices")
        if choices is None:
            continue

        delta = choices[0].get("delta")

        # TODO rewrite this to be more robust streaming json parser
        payload += delta.get("function_call", {}).get("arguments", "")

        if is_parent_list:
            fn_name = camelcase_to_words(name).replace(" ", "_").lower()
            s = extract_partial_list(payload, key=fn_name)
            if s is None or len(s) == 0 or len(s) <= count:
                continue

            for data in s[count:]:
                yield output_type(**remove_nas(data))

            count = len(s)


async def parallelize_streaming_chat_wrapper(
    num_workers,
    messages,
    functions,
    model,
    config,
    temperature,
    output_type,
    is_parent_list,
    name,
):
    print(f"parallelize_streaming_chat_wrapper {num_workers}")

    async def collect_results(generator):
        results = []
        async for item in generator:
            results.append(item)
        return results

    tasks = [
        asyncio.ensure_future(
            collect_results(
                streaming_chat_wrapper(
                    messages,
                    functions,
                    model,
                    config,
                    temperature,
                    output_type,
                    is_parent_list,
                    name,
                )
            )
        )
        for _ in range(num_workers)
    ]

    for future in asyncio.as_completed(tasks):
        results = await future
        for result in results:
            yield result


async def _brainstorm(
    text: str,
    n: int = 10,
    temperature: float = 0.2,
    output_type: type[BaseModel] | BaseModel | str | None = None,
    positive_examples: list[str | BaseModel] = None,
    negative_examples: list[str | BaseModel] = None,
    batch_size: int = 10,
    max_workers: int = 5,
    model: str = GPT_MODEL,
    config: Config = None,
) -> list[BaseModel] | list[str]:
    """Generate ideas using a text prompt"""
    config = config or Config()
    positive_examples = positive_examples or []
    negative_examples = negative_examples or []
    output_type = output_type or str

    if is_parent_list := str(output_type).lower().startswith("list"):
        assert (
            len(output_type.__args__) == 1
        ), "Only one type argument is allowed for list types"
        output_type = output_type.__args__[0]
    assert issubclass(output_type, BaseModel), "output_type must be a Pydantic model"

    schema = output_type.model_json_schema()
    name = schema["title"]

    batch_size = min(batch_size, n)

    messages = create_messages(
        text,
        name=name,
        labels=schema["properties"],
        n=batch_size,
        positive_examples=positive_examples,
        negative_examples=negative_examples,
    )
    functions = create_functions(
        is_parent_list,
        schema,
        fn_name=camelcase_to_words(name).replace(" ", "_").lower(),
    )

    if n % batch_size != 0:
        max_workers = max(1, min(max_workers, 1 + n // batch_size))
    else:
        max_workers = max(1, min(max_workers, n // batch_size))

    yield_count = 0
    if max_workers == 1:
        results = streaming_chat_wrapper(
            messages,
            functions,
            model,
            config,
            temperature,
            output_type,
            is_parent_list,
            name,
        )
    else:
        results = parallelize_streaming_chat_wrapper(
            max_workers,
            messages,
            functions,
            model,
            config,
            temperature,
            output_type,
            is_parent_list,
            name,
        )
    async for result in results:
        yield result
        yield_count += 1
        if yield_count >= n:
            break


async def brainstorm(
    text: str,
    n: int = 10,
    temperature: float = 0.2,
    output_type: type[BaseModel] | BaseModel | str | None = None,
    positive_examples: list[str | BaseModel] = None,
    negative_examples: list[str | BaseModel] = None,
    batch_size: int = 10,
    max_workers: int = 5,
    model: str = GPT_MODEL,
    config: Config = None,
) -> list[BaseModel] | list[str]:
    yield_count = 0
    while yield_count < n:
        new_yield_count = 0
        async for result in _brainstorm(
            text,
            n=n - yield_count,
            temperature=temperature,
            output_type=output_type,
            positive_examples=positive_examples,
            negative_examples=negative_examples,
            batch_size=batch_size,
            max_workers=max_workers,
            model=model,
            config=config,
        ):
            yield result
            new_yield_count += 1
            yield_count += 1
            if yield_count >= n:
                break
        if new_yield_count == 0:
            print("No more results")
            break


class BusinessIdea(BaseModel):
    """A business idea generated using the Jobs-to-be-done framework
    For example "We help [adj] [target_audience] do [action] so they can [benefit or do something else]"
    """

    target_audience: str = Field(title="Target Audience")
    action: str = Field(title="Action")
    benefit: str = Field(title="Benefit or next action")
    adj: str | None = Field(
        title="Adjective",
        description="Optional adjective describing the target audience's condition",
    )
